/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * 	This is the firestorm config file parser, it's a two
 * 	phase thing. Read in all the config lines, then only
 * 	apply the configuration after everything has been read.
 * 	It's fairly easy to add new config items, see chandlers[]
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <packet.h>
#include <cleanup.h>
#include <capture.h>
#include <preproc.h>
#include <loader.h>
#include <alert.h>
#include <target.h>
#include <decode.h>
#include <parser.h>
#include <signature.h>
#include <realm.h>
#include <preproc.h>
#include <capdev.h>
#include <matcher.h>
#include <template.h>
#include <alert.h>
#include <conf.h>

#define _BSD_SOURCE
#include <grp.h>

#define CONF_MAXLEN 1024

uid_t init_euid;

/* order here is crucial */
struct conf_handler chandlers[]={
	{"firestorm_root",	0,           NULL, NULL, root_apply},
	{"chroot",		0,           NULL, NULL, chroot_apply},
	{"logfile",		0,           NULL, NULL, log_apply},
	{"load_plugins",	CHF_ITERATE, NULL, NULL, lps_apply},
	{"load_plugin",		CHF_ITERATE, NULL, NULL, lpf_apply},
	{"",			CHF_HIDDEN,  NULL, NULL, done_plugins},
	{"require",		CHF_ITERATE, NULL, NULL, req_apply},
	{"capture",		0,           NULL, NULL, cap_apply},
	{"effective_gid",	0,           NULL, NULL, gid_apply},
	{"effective_uid",	0,           NULL, NULL, uid_apply},
	{"load_db",		CHF_ITERATE, NULL, NULL, NULL},
	{"preprocessor",	CHF_ITERATE, NULL, NULL, preproc_apply},
	{"output",		CHF_ITERATE, NULL, NULL, output_apply},
	{"signatures",		CHF_ITERATE, NULL, NULL, sig_apply},
	{"",			CHF_HIDDEN,  NULL, NULL, sig_count},
	{NULL, 0, NULL, NULL, NULL}
};

typedef void *(*proc_req_func)(char *);
struct {
	char		*item;
	proc_req_func	fn;
}reqs[]={
	{"proto",	(proc_req_func)decode_proto},
	{"realm",	(proc_req_func)realm_find},
	{"preproc",	(proc_req_func)preproc_find}, /* not needed */
	{"capdev",	(proc_req_func)capdev_find}, /* not needed */
	{"target",	(proc_req_func)target_find}, /* not needed */
	{"matcher",	(proc_req_func)matcher_find}, /* not needed */
	{"plugin",	(proc_req_func)loader_find}, /* not really needed */
	{NULL, NULL}
};


/* Find a config handler */
struct conf_handler *conf_find(char *name)
{
	int i;

	for(i=0; chandlers[i].name!=NULL; i++) {
		if ( !(chandlers[i].flags & CHF_HIDDEN) && 
			!strcmp(name, chandlers[i].name) ) {
			return &chandlers[i];
		}
	}
	
	return NULL;
}


/* Parse and process a single line */
int conf_line(char *fn, int ln, char *line)
{
	char *keyword, *args=NULL;
	char *cur;
	int state=0;
	struct conf_handler *cnf;
	struct conf_list *cl;
	
	/* Little state machine to parse out the
	 * keyword from the rest of the stuff... */
	for(cur=keyword=line; ;cur++) {
		if ( state==0 ) {
			if ( *cur==' ' || *cur=='\t' ) {
				state=1;
				*cur='\0';
				continue;
			}
		}else if ( state==1 ) {
			if ( *cur!=' ' && *cur!='\t' &&
				*cur!='\r' && *cur!='\n') {
				args=cur;
				state=2;
				continue;
			}
		}

		if ( *cur=='\r' || *cur=='\n' ) {
			*cur='\0';
			break;
		}
	}

	if ( !(cnf=conf_find(keyword)) ) {
		cleanup(EXIT_ERR,
			"Unknown keyword '%s' in %s:%i",
			keyword, fn, ln);
	}

	if ( !(cnf->flags & CHF_ITERATE) && cnf->list ) {
		cleanup(EXIT_ERR,
			"Using '%s' again on %s:%i",
			keyword, fn, ln);
	}

	/* Add the conf_list to the relevent list */

	if ( !(cl=malloc(sizeof(*cl))) ) {
		cperror("malloc");
	}

	if ( args && !(cl->str=strdup(args)) ) {
		cperror("strdup");
	}else if ( !args ) {
		cl->str=NULL;
	}

	cl->line=ln;
	cl->filename=fn;

	cl->next=NULL;
	if ( !cnf->list ) {
		cnf->list=cl;
		cnf->end=cl;
	}else{
		cnf->end->next=cl;
		cnf->end=cl;
	}

	return 1;
}

/* Parse a config file */
void conf_load(char *fn)
{
	FILE *f;
	int line=0;
	static char buf[CONF_MAXLEN];

	/* Not likely! */
	if ( !fn ) return;

	if ( (f=fopen(fn, "r"))==NULL ) {
		cleanup(EXIT_ERR, "%s: %s", fn, get_err());
	}
	
	while(fgets(buf, CONF_MAXLEN, f)) {
		char *p=buf;
		line++;

		/* Strip leading whitespace */
		while(*p && (*p=='\t' || *p==' ') ) p++;

		/* Check it aint empty line or comment */
		if ( *p=='\0' || *p=='\n' || *p=='\r' || *p=='#' ) continue;

		if ( !conf_line(fn, line, p) ) {
			fclose(f);
			goto err;
		}
	}

	fclose(f);
	return;

err:
	cleanup(EXIT_ERR, "config: Parse error in %s:%i", fn, line);
}

/* Iterate through all items in a config list
 * obliterating everything as we go... */
void conf_iterate(struct conf_handler *h)
{
	struct conf_list *foo, *bar;

	for(foo=h->list; foo;) {
		bar=foo;
		foo=foo->next;

		h->apply(h->name,bar->str);

		free(bar->str);
		free(bar);
	}
}

/* Apply all the config */
void conf_apply(void)
{
	int i;

	init_euid=geteuid();
	
	for(i=0; chandlers[i].name!=NULL; i++) {
		if ( !chandlers[i].apply ) continue;
		
		if ( chandlers[i].flags & CHF_ITERATE ) {
			conf_iterate(&chandlers[i]);
		}else{
			char *args=NULL;

			if ( chandlers[i].list )
				args=chandlers[i].list->str;
			
			chandlers[i].apply(chandlers[i].name, args);

			if ( chandlers[i].list ) {
				free(chandlers[i].list->str);
				free(chandlers[i].list);
			}
		}
	}
}

/* We need to resolve decoder dependancies right after
 * the plugins are loaded */
void done_plugins(char *k, char *a)
{
	decode_resolve_all();
}

/* load plugins by dir */
void lps_apply(char *k, char *a)
{
	if ( !a ) return;
	
	loader_load_dir(a);
}

/* Load plugins by file */
void lpf_apply(char *k, char *a)
{
	if ( !a ) return;

	if ( !loader_load(a) ) 
		cleanup(EXIT_ERR, "Couldn't load plugin '%s'", a);
}

void sig_count(char *k, char *a)
{
	signature_stats();
}

void sig_apply(char *k, char *a)
{
	char *type, *path;

	conf_parse_arg(a, &type, &path);

	if ( !type || !path ) {
		return;
	}

	if ( !parser_load(type, path) ) {
		cleanup(EXIT_ERR, "Unable to load %s signatures (%s)",
			type, path);
	}
}

void preproc_apply(char *k, char *a)
{
	char *type, *args;

	conf_parse_arg(a, &type, &args);

	if ( !type ) return;

	if ( !preproc_load(type, args) ) {
		cleanup(EXIT_ERR, "Unable to add '%s' preprocessor", type);
	}
}

void cap_apply(char *k, char *a)
{
	char *type, *interface;

	conf_parse_arg(a, &type, &interface);

	if ( !type ) return;
	
	if ( !capture_add(type, interface) ) {
		cleanup(EXIT_ERR, "Unable to add '%s' capture (%s)",
			type, interface);
	}

}

void conf_parse_arg(char *str, char **arg, char **rest)
{
	int state=0;

	if ( !str ) {
		*arg=NULL;
		*rest=NULL;
		return;
	}

	*rest=NULL;
	
	for(*arg=str; *str; str++) {
		if ( state==0 ) {
			if ( *str==' ' || *str=='\t' ) {
				state=1;
				*str='\0';
				continue;
			}
		}else if ( state==1 ) {
			if ( *str!=' ' && *str!='\t' ) {
				*rest=str;
				break;
			}
		}
	}
}

void root_apply(char *k, char *a)
{
	char *rootdir="/";
	
	if ( a ) rootdir=a;

	if ( chdir(rootdir) ) {
		cleanup(EXIT_ERR, "%s: chdir(): %s", rootdir, get_err());
	}
}

void chroot_apply(char *k, char *a)
{
	char *ch="yes";
	
	if ( a ) ch=a;
	
	if ( !init_euid && !strcmp("yes", ch) && chroot(".") ) {
		cleanup(EXIT_ERR, "chroot(): %s", get_err());
	}
}

void uid_apply(char *k, char *a)
{
	uid_t userid;
	
	if ( !a ) return;

	/* Only drop root if we ARE root! */
	if ( init_euid ) return;

	if ( !strtouint(a, &userid) ) {
		cleanup(EXIT_ERR, "Invalid uid: %s", a);
	}

	/* Ignore zero */
	if ( !userid ) return;
		
	if ( setuid(userid) == -1 ) {
		cleanup(EXIT_ERR, "Unable to drop uid to %s: %s",
			a, get_err());
	}
}

void req_apply(char *k, char *a)
{
	char *func=NULL, *val=NULL;
	int i;

	conf_parse_arg(a, &func, &val);

	if ( !func || !val ) return;

	for(i=0; reqs[i].item; i++) {
		if ( !strcmp(func, reqs[i].item) ) {
			if ( reqs[i].fn(val) ) return;
		}
	}

	cleanup(EXIT_ERR, "Unable to meet requirement: "
		"%s -> \"%s\"\n", func, val);
}

void gid_apply(char *k, char *a)
{
	gid_t groupid;

	/* Initialise all captures before
	 * dropping root privs */
	capture_init_all();

	if ( !a ) return;

	/* Only drop root if we ARE root! */
	if ( init_euid ) return; 

	if ( !strtouint(a, &groupid) ) {
		cleanup(EXIT_ERR, "Invalid gid: %s", a);
	}

	if ( !groupid ) return;

	if ( setgroups(1, &groupid) == -1 ||
		setgid(groupid) == -1 ) {
		cleanup(EXIT_ERR, "Unable to drop gid to %s: %s",
			a, get_err());
	}
}

void output_apply(char *k, char *a)
{
	char *gen, *type, *args;
	
	conf_parse_arg(a, &gen, &args);
	if ( !args ) {
		cleanup(EXIT_ERR, "Unable to parse output "
			"line for '%s'", gen);
	}
	
	conf_parse_arg(args, &type, &args);
	
	if ( !alert_output(gen, type, args) ) {
		cleanup(EXIT_ERR, "Unable to register output "
			"module: %s -> %s(%s)\n", gen, type, args);
	}
}

void log_apply(char *k, char *a)
{
	int fd, fds;

	/* Close all fds apart from (stdin,stdout,stderr) */
	fds=sysconf(_SC_OPEN_MAX);
	while ( fds>2 ) close(fds--);
	
	/* Stay in the foreground */
	if ( !a ) {
		dup2(1, 0); 
		dup2(1, 2);
		return;
	}

	/* Open up the logfile, don't forget permissions! */
	if ( (fd=open(a, O_WRONLY|O_CREAT|O_TRUNC, 0600))<0 ) {
		printf("conf: %s: open(): %s\n", a, get_err());
		cleanup(EXIT_ERR, "Unable to open logfile ('%s')", a);
	}

	/* Make stdout be the logfile */
	if ( fd != 1 ) {
		if ( dup2(fd, 1)<0 ) {
			cleanup(EXIT_ERR,
				"Unable to dup2() logfile: %s",
				get_err());
		}

		/* Don't waste descriptors */
		close(fd);
	}

	dup2(1, 0); /* Make stdin == stdout, EBADF for reading */
	dup2(1, 2); /* Make stderr == stdout */

	/* Good old fashioned fork()/setsid()/fork()
	 * means we can never again become part of a
	 * job control group or regain a controlling
	 * tty - thank god */
	switch(fork()) {
		case 0: break;
		case -1:
			cleanup(EXIT_ERR,
				"Unable to fork(): %s",
				get_err());
		default: _exit(0);
	}

	if ( setsid()<0 ) {
		cleanup(EXIT_ERR,
			"Unable to setsid(): %s",
			get_err());
	}

	switch(fork()) {
		case 0: break;
		case -1:
			cleanup(EXIT_ERR,
				"Unable to fork() again: %s",
				get_err());
		default: _exit(0);
	}
}
